"""This module contains tools for dimensionaly reduction and geomerty."""

import numpy as np
from typing import Optional, Callable, List
import copy
from matplotlib import pyplot as plt


class AffineSubspace:
    """This object contains data defining an affine linear subspace of some higher
      dimensional real or complex vector space.
    It consists of an orthonormal basis of directions
    and an arbitary vector as the center.

    Attribues:

        directions: An approximate orthonormal set of vectors from some high
        dimensional superspace.
        Must be of shape (number of directions, dimension of superspace)

        center: A vector from the same space as the directions.

        sharpness: The decimals error allowed in the check for orthonormality.
        That means both in terms of norm and orthogonality of each pair of directions.

    """

    def __init__(
        self,
        directions: np.ndarray,
        center: Optional[np.ndarray] = None,
        sharpness: Optional[int] = 2,
        orthonormalize: Optional[bool] = True,
    ):
        d_num = directions.shape[0]
        d_dim = directions.shape[1]

        self.sharpness = sharpness

        # checking orthonormality and enforcing it if necessary
        if (
            orthonormalize is False
            or (
                np.round(np.dot(directions, np.transpose(directions)), sharpness)
                == np.identity(d_num, dtype=float)
            ).all()
        ):
            self.directions = directions
        else:
            # warnings.warn(
            #     """Directions were not given as orthonormal basis so they will be
            #     adjusted automatically."""
            # )
            self.directions = gramschmidt(directions.transpose()).transpose()
            if not (
                np.round(
                    np.dot(self.directions, np.transpose(self.directions)), sharpness
                )
                == np.identity(d_num, dtype=float)
            ).all():
                raise AttributeError(
                    "Provided directions are not linearly independent."
                )

        # checking center dimension, or setting it to 0 if none was provided
        if center is None:
            self.center = np.zeros((1, d_dim))
        else:
            if center.shape == (1, d_dim):
                self.center = center
            else:
                raise AttributeError(
                    "Center dimension must be the same as direction dimension."
                )

    def orth(self):
        """This method returns the orthogonal complement of the AffineSubspace.
        Note that this forgets everything about the orientation of the directions,
        in particular, this operation is not self inverse regarding the exact basis
        """
        d_dim = self.directions.shape[1]

        candidates = subspace_projection(
            pointcloud=np.eye(d_dim) + np.tile(self.center, [d_dim, 1]),
            target_space=self,
            orthogonal=True,
            relative=True,
        )

        i_c = 0
        found_start = False
        while not found_start:
            if np.linalg.norm(candidates[[i_c], :]) < (
                1 / np.power(10, self.sharpness)
            ):
                i_c += 1
            else:
                OC = AffineSubspace(
                    candidates[[i_c], :], center=self.center, sharpness=self.sharpness
                )
                found_start = True

        while (
            not OC.directions.shape[0]
            == self.directions.shape[1] - self.directions.shape[0]
        ):
            i_c += 1

            if i_c == d_dim:
                raise AttributeError(
                    "Did not find enough linear independent candidates."
                )

            if not np.linalg.norm(
                subspace_projection(
                    pointcloud=candidates[[i_c], :] + self.center,
                    target_space=OC,
                    orthogonal=True,
                    relative=True,
                )
            ) < (1 / np.power(10, self.sharpness)):
                OC = AffineSubspace(
                    np.append(OC.directions, candidates[[i_c], :], axis=0),
                    center=OC.center,
                    sharpness=OC.sharpness,
                )

        if (
            not OC.directions.shape[0]
            == self.directions.shape[1] - self.directions.shape[0]
        ):
            raise AttributeError("Dimension of complement is not as expected.")
        else:
            return OC


class Hessian:
    """This is a class for hessian matrices evaluated at specific points.

    Attributes:

        matrix: the hessian matrix itself

        func: the function for which it was calculated

        subspace: the point at which it was calculated and
         the directions in which the partial derivatives were taken

        epsilon: the offset used in the numeric calculation

    Methods:

        calc_evs: calculates eigenvectors and eigenvalues of the hessian matrix

        show_evs: creates a scatterplot of the eigenvalues, lowest to highest
    """

    def __init__(
        self,
        matrix: np.ndarray,
        func: Optional[Callable[[np.ndarray], float]] = None,
        subspace: Optional[AffineSubspace] = None,
        epsilon: Optional[float] = None,
    ):
        self.matrix = matrix
        self.func = func
        self.subspace = subspace
        self.epsilon = epsilon

    def calc_evs(
        self,
    ):
        """Calculates and adds the eigenvalues and eigenvectors
        as attributes to the object. Sorts eigenvalues lowest to highest."""
        if not hasattr(self, "eigenvalues"):
            ev = np.linalg.eigh(self.matrix)

            # sorting the eigenvalues in ascending order
            order = np.argsort(ev.eigenvalues)

            self.eigenvalues = ev.eigenvalues[order]
            self.eigenvectors = ev.eigenvectors[:, order]

    def show_evs(self, **plot_kwargs):
        """Creates a scatterplot of the eigenvalues. y-axis is the actual eigenvalue,
        x-axis is the index of each eigenvalue in the order lowest to highest"""
        if not hasattr(self, "eigenvalues"):
            self.calc_evs()

        plt.scatter(range(1, self.matrix.shape[0] + 1), self.eigenvalues, **plot_kwargs)
        plt.xticks(
            ticks=range(1, self.matrix.shape[0] + 1),
            labels=range(1, self.matrix.shape[0] + 1),
        )
        plt.xlabel("index")
        plt.ylabel("eigenvalue")


def subspace_projection(
    pointcloud: np.ndarray,
    target_space: AffineSubspace,
    orthogonal: Optional[bool] = False,
    relative: Optional[bool] = False,
) -> np.ndarray:
    """This function takes as input a pointcloud and an affine linear subspace
    and puts out the projection of the pointcloud onto that subspace, or its orthogonal
    complement.

    Please note that I believe this is the most accurate way to to this in numpy but it
    is still not perfect. Especially when you involve irraltional entries
    like np.sqrt(2) or np.pi you will likely get some inaccurary.

    IMPORTANTLY it uses the fact that the directions of AffineSubspace are orthonormal.

    Input:

        - pointcloud: an arbitrary collection with same dimension as the superspace,
            ie shape (number of points, dimension of superspace)
        - target_space: the subspace that defines the projection
        - orthogonal: whether to project to target_space (False) or
            its orthogonal complement (True)
        - relative: whether to represent the output relative
            to zero (False) or to the center of the AffineSubspace (True)

    Output:

        - np.ndarray with same shape as pointcloud, representing its projection

    """

    if pointcloud.shape[1] != target_space.directions.shape[1]:
        raise AttributeError(
            """The shape of pointclould must be (number of points, dimension of
            superspace) even if its just one point."""
        )

    d_num = target_space.directions.shape[0]
    p_num = pointcloud.shape[0]
    points_relative = pointcloud - np.repeat(target_space.center, p_num, axis=0)

    # calculating the coefficients that each element of the subspace basis
    # would have in a linear combination for each point,
    # if it was extended to a full orthonormal basis of the space
    coeff = np.zeros([d_num, p_num])
    for i_dir in range(d_num):
        coeff[i_dir] = np.dot(
            points_relative, target_space.directions[i_dir, :].transpose()
        )

    # returning the relevant "parts" of the points, possibly offset by the center
    if orthogonal is False:
        return np.dot(target_space.directions.transpose(), coeff).transpose() + (
            not relative
        ) * np.repeat(target_space.center, p_num, axis=0)
    elif orthogonal is True:
        return (
            points_relative
            - np.dot(target_space.directions.transpose(), coeff).transpose()
            + (not relative) * np.repeat(target_space.center, p_num, axis=0)
        )


def gramschmidt(V: np.ndarray):
    """This function just orthogonalizes a matrix V using the Gram-Schmidt algorithm.
    It turns the columns of V into an orthonormal basis/set of vectors.
    It is here because other methods, like scipy.linalg.orth
    seem to be less accurate.
    """

    U = np.zeros(shape=V.shape, dtype=float)
    U[:, 0] = V[:, 0] / np.linalg.norm(V[:, 0])
    for i in range(1, V.shape[1]):
        U[:, i] = V[:, i]
        for j in range(i):
            U[:, i] = U[:, i] - np.dot(U[:, j].transpose(), U[:, i]) * U[:, j]
        U[:, i] = U[:, i] / np.linalg.norm(U[:, i])
    return U


def numeric_partial_derivative(
    func: Callable[[np.ndarray], float],
    pos: List[int],
    point: np.ndarray,
    epsilon: Optional[float] = 0.01,
) -> float:
    """This function approximates chains of partial derivatives using central finite
    differences around a specific point.

    Input:
        - func: the function to be differentiated
        - pos: the positions for which the partial derivatives are to be taken,
            ie the first entry of this list would be the rightmost index
            when writing it on paper
            (the entries must be between 1 and the dimension of the argument space)
        - point: the point at which the derivative is to be evaluated
        - epsilon: the offset used to calculate the numeric derivative,
            should either be very small or chosen according to some parameter shift rule
    """
    depth = len(pos)

    value_acc = 0
    # going through the binary representations of all numbers up to 2^depth
    # this encodes the nested additions and substractions in the approximation
    for i in range(np.power(2, depth)):
        i_c = copy.copy(i)
        par = copy.copy(point)
        sign = 1

        # depending on current digit, shift the parameter vector by foward or backward
        for i_d in range(depth - 1, -1, -1):

            if np.power(2, i_d) <= i_c:
                i_c = i_c - np.power(2, i_d)
                esign = -1
            else:
                esign = 1

            sign = sign * esign
            par[pos[i_d]] = par[pos[i_d]] + esign * epsilon / 2

        value_acc = value_acc + sign * func(par)

    # returning the accumulated sum of differences
    # divided by the offset for each step of derivation
    return value_acc / np.power(epsilon, depth)


def numeric_gradient(
    func: Callable[[np.ndarray], float],
    subspace: AffineSubspace,
    epsilon: Optional[float] = 0.01,
    relative: bool = False,
) -> np.ndarray:
    """This approximates the gradient of a function
    (restricted to the given subspace) unsing finite central differences.

    Input:
        - func: any function mapping an array of real numbers to a real number
        - subspace: an affine linear subspace of the parameter space,
            determines around which point to calculate the gradient
            and what basis to represent it in
        - epsilon: the offset used in the approximation, should either be small
            or chosen according to some parameter shift rule
    """

    d_num = subspace.directions.shape[0]
    grad = np.zeros((1, d_num))

    # reparametrizing the function to the subspace
    def func_sub(
        par=np.ndarray,
    ) -> float:
        return func(subspace.center + np.dot(subspace.directions.transpose(), par))

    for k in range(d_num):
        grad[0, k] = numeric_partial_derivative(
            func_sub, [k], np.zeros([d_num]), epsilon=epsilon
        )

    if relative is False:
        return grad
    else:
        return np.dot(grad, subspace.directions)


def numeric_hessian(
    func: Callable[[np.ndarray], float],
    subspace: AffineSubspace,
    epsilon: Optional[float] = 0.01,
) -> Hessian:
    """This approximates the hessian of a given function
    (restricted to the given subspace) unsing finite central differences.

    Input:
        - func: any function mapping an array of real numbers to a real number
        - subspace: an affine linear subspace of the parameter space,
            determines around which point to calculate the hessian
            and in which directions to take the partial derivatives
        - epsilon: the offset used in the approximation, should either be small
            or chosen according to some parameter shift rule
    """
    d_num = subspace.directions.shape[0]
    H = np.zeros([d_num, d_num])

    # reparametrising the function to the subspace
    def func_sub(
        par=np.ndarray,
    ) -> float:
        return func(subspace.center + np.dot(subspace.directions.transpose(), par))

    # getting all second order partial derivatives
    for j in range(d_num):
        for k in range(d_num):
            H[j, k] = numeric_partial_derivative(
                func_sub, [j, k], np.zeros([d_num]), epsilon=epsilon
            )

    return Hessian(matrix=H, func=func, subspace=subspace, epsilon=epsilon)



"""This module collects methods that scan a landscape along affine linear subspaces,
so that they can easily be plotted."""

from typing import Callable, List, Optional, Union

import numpy as np
from matplotlib import pyplot as plt

from .dr_tools import AffineSubspace, numeric_hessian


class LinearScan:
    """This object contains the results of a linear scan.

    Attributes:

        result: a grid of values sampled from a real valued function.

        subspace: the affine linear subspace along which the samples were taken.

        scope: tells you how far the subspace was scanned in each direction.
            shape is (number of directions, 2)

        func: optional attribute to record the function that was sampled.

    Methods:

        show: Plots the result of the scan, if it is 1 or 2 dimensional.

    """

    def __init__(
        self,
        result: np.ndarray,
        subspace: AffineSubspace,
        scope: np.ndarray,
        func: Optional[Callable[[np.ndarray], float]] = None,
    ):
        self.result = result
        self.subspace = subspace
        self.scope = scope
        self.func = func

    def show(self, **plot_kwargs) -> None:
        """This method plots the results of the scan, unless it has more than 2
        dimensions."""

        # 2D scan
        if self.result.ndim == 2:
            plt.imshow(
                np.transpose(self.result),
                extent=[
                    self.scope[0, 0],
                    self.scope[0, 1],
                    self.scope[1, 0],
                    self.scope[1, 1],
                ],
                **plot_kwargs,
            )
            plt.xlabel("first scan direction")
            plt.ylabel("second scan direction")

        # 1D scan
        elif self.result.ndim == 1:
            y = self.result
            x = np.linspace(
                self.scope[0, 0], self.scope[0, 1], len(self.result)
            )
            plt.plot(x, y, **plot_kwargs)
            plt.xlabel("scan direction")
            plt.ylabel("function value")

        # error for too many dimensions
        else:
            raise AttributeError(
                "Only scans of dimension 1 or 2 can be shown."
            )


class ScanCollection:
    """This object contains a collection of Linear scan objects.

    Attributes:

        scanlist: A list of LinearScan objects

    Method:

        show: plots all of the scans at once

    """

    def __init__(
        self,
        scanlist: List[LinearScan],
    ):
        self.scanlist = scanlist

    def show(
        self, showlist: Optional[List[int]] = None, **plot_kwargs
    ) -> None:
        """This method plots all the scans of the collection in one figure.
        All of them must be 1D for this to work.

        Input:

            showlist: The idices of the scans that are to be shown.
                For example [0,2] would show only the first and third scan.

        To display the legend execute matplotlib.pyplot.legend() after this method.
        """

        if showlist is None:
            showlist = range(len(self.scanlist))

        if self.scanlist[0].subspace.directions.shape[0] == 1:
            for i_scan in showlist:
                _plot_kwargs = {
                    **{"label": "direction " + str(i_scan)},
                    **plot_kwargs,
                }
                self.scanlist[i_scan].show(**_plot_kwargs)
        else:
            raise AttributeError(
                "Only scans of dimension 1 can be shown as collectives."
            )


class PathScan:
    """This object contains a 1D scan that is related to a path in the landscape.

    Attributes:

        path: A sequence of points in the landscape. (shape [num_points, dim])
            For example this could be a training trajectory.

        result: A flat array of loss values.

        mode: The method used to obtain the results. There should be the following:
            - raw (just every node of the path evaluated)
            - compressed (every node of the path evaluated given a minimal distance to the last evaluated point)
            - refined (every node of the path and some number of nodes in between two steps)
            - segmented (some number of nodes, equidistant on the path, probably wont include most nodes of the path)

        function: the function that was scanned
    """

    def __init__(
        self,
        result: np.ndarray,
        path: np.ndarray,
        mode: str,
        func: Optional[Callable[[np.ndarray], float]] = None,
    ):
        self.result = result
        self.path = path
        self.mode = mode
        self.func = func

    def show(self, **plot_kwargs) -> None:
        plt.plot(self.result, **plot_kwargs)


def landscape_scan_linear(
    func: Callable[[np.ndarray], float],
    subspace: AffineSubspace,
    scope: Union[np.ndarray, float] = 5,
    resolution: Optional[Union[int, np.ndarray]] = 10,
    pools: Optional[int] = 1,
) -> LinearScan:
    """This function scans a portion of an affine linear subspace of the higher
    dimensional lanscape created by func, with a specified resolution.

    Input:

        func: The function that defines the landscape.

        subspace: The affine linear subspace along which the scan is performed.

        scope: How far to scan in each direction of subspace.
        The size has to be (d_num,2), where for each direction the first entry is the
        beginning of the scope and the second is the end.
        If provided as float, the beginning will be -scope and the end +scope.

        resolution: How many samples to take in each direction of subspace.
        If provided as int, the resolution will be the same for each direction.

        pools: the number of pools to be used in the parallelization of the
        evaluation of func on the parameter grid. This is currently deactivated
        as it might not work as intended.

    Output:

        A LinearScan object.

    Notes:

        The total number of func calls is prod(resolution).

        The outermost values sampled are defined by scope.
        In particular the exact center of the subspace will only be sampled if all
        resolutions are odd.

    """

    # initalizing some important variables:

    d_num = subspace.directions.shape[0]
    d_dim = subspace.directions.shape[1]

    if not isinstance(scope, np.ndarray):
        scope = scope * np.append(-np.ones([d_num, 1]), np.ones([d_num, 1]), 1)

    if not (scope[:, 0] < scope[:, 1]).all():
        raise ValueError(
            """scope[id_d,0] must be strictly smaller than scope[id_d,1] for each
            direction."""
        )

    if np.isscalar(resolution):
        resolution = resolution * np.ones(d_num, dtype=int)

    # creating the grid of arguments to be scanned:
    # by first initializing every point as bottom_corner and then adding scaled grids
    # of all the direction vectors
    bottom_corner = (
        subspace.center.flatten()
        + np.sum(
            np.dot(np.diag(scope[:, 0]), subspace.directions), 0
        ).flatten()
    )
    grid = np.tile(bottom_corner, np.append(resolution, [1]))
    for id_d in range(d_num):
        steps_d = np.linspace(
            0, scope[id_d, 1] - scope[id_d, 0], resolution[id_d]
        )
        tile_d = np.multiply.outer(steps_d, subspace.directions[id_d, :])

        morph_d = np.ones(grid.ndim, dtype=np.int64)
        morph_d[-1] = d_dim
        morph_d[id_d] = resolution[id_d]
        tile_d = tile_d.reshape(morph_d)

        res_d = np.append(resolution, [1])
        res_d[id_d] = 1

        grid_d = np.tile(tile_d, res_d)

        grid = grid + grid_d

    # applying the function to all the values in the grid to get the scan:
    # this should be the most expensive step

    # used some tricks with reshape to write a paralellization friendly loop that
    # iterates over all the argument vectors in the grid

    res_total = np.prod(resolution)
    result = np.zeros(res_total)
    grid_r = grid.reshape(res_total, d_dim)

    # # parallelized (experimental)
    # with mp.Pool(pools) as p:
    #     result = p.map(func, grid_r)
    # result = np.array(result).reshape(resolution)

    # unparallelized
    for id_r in range(res_total):
        result[id_r] = func(grid_r[[id_r], :])
    result = np.array(result).reshape(resolution)

    return LinearScan(result=result, subspace=subspace, scope=scope)


def collective_scan_linear(
    func: Callable[[np.ndarray], float],
    subspace: AffineSubspace,
    scope: Union[np.ndarray, float] = 5,
    resolution: Optional[Union[int, np.ndarray]] = 10,
    pools: Optional[int] = 1,
) -> ScanCollection:
    """This functions creates a 1D scan
    in every direction of the provided subspace.

    Input:

        func: The function that defines the landscape.

        subspace: The affine linear subspace along which the scan is performed.

        scope: How far to scan in each direction of subspace.
            The size has to be (d_num,2), where for each direction the first entry
            is the beginning of the scope and the second is the end.
            If provided as float, the beginning will be -scope and the end +scope.
            resolution: How many samples to take in each direction of subspace.
            The total number of func calls is prod(resolution).

    Output:
        ScanCollection object that contains 1D scans for each direction of the subspace.

    """
    d_num = subspace.directions.shape[0]

    if not isinstance(scope, np.ndarray):
        scope = scope * np.append(-np.ones([d_num, 1]), np.ones([d_num, 1]), 1)

    if not (scope[:, 0] < scope[:, 1]).all():
        raise ValueError(
            """scope[id_d,0] must be strictly smaller
            than scope[id_d,1] for each direction."""
        )

    if isinstance(resolution, int):
        resolution = resolution * np.ones(d_num, dtype=int)

    # defining and immediately scanning all the 1D subspaces
    scanlist = []
    for i_scan in range(d_num):
        scanlist.append(
            landscape_scan_linear(
                func,
                AffineSubspace(
                    directions=subspace.directions[[i_scan], :],
                    center=subspace.center,
                ),
                scope=scope[[i_scan], :],
                resolution=resolution[i_scan],
                pools=pools,
            )
        )

    return ScanCollection(scanlist)


def hessian_scan(
    func: Callable[[np.ndarray], float],
    subspace: AffineSubspace,
    scope: Optional[Union[np.ndarray, float]] = 5,
    resolution: Optional[Union[int, np.ndarray]] = 10,
    epsilon: Optional[float] = 0.01,
    pools: Optional[int] = 1,
):
    """This function calculates the Hessian of func at the center of subspace
    and performs a collective scan in the directions of the eigenvectors.

    Input:

        func: The function that defines the landscape.

        subspace: The affine linear subspace along which the scan is performed.

        scope: How far to scan in each direction of subspace.
        The size has to be (d_num,2), where for each direction the first entry is the
        beginning of the scope and the second is the end.
        If provided as float, the beginning will be -scope and the end +scope.

        resolution: How many samples to take in each direction of subspace.
        If provided as int, the resolution will be the same for each direction.

    Output:

        The Hessian and a ScanCollection object.

    """

    H = numeric_hessian(func=func, subspace=subspace, epsilon=epsilon)
    H.calc_evs()
    subspace_H = AffineSubspace(
        directions=np.dot(H.eigenvectors.transpose(), subspace.directions),
        center=subspace.center,
    )

    scan_H = collective_scan_linear(
        func=func,
        subspace=subspace_H,
        scope=scope,
        resolution=resolution,
        pools=pools,
    )

    return scan_H, H


def pathscan(
    func: Callable[[np.ndarray], float],
    path: np.ndarray,
    mode: str,
    resolution: Optional[int] = 0,
    stepsize: Optional[float] = 0,
    pools: Optional[int] = 1,
) -> PathScan:
    l_p = path.shape[0]
    dim = path.shape[1]

    if mode == "raw":
        scanpath = path
    if mode == "refined":
        scanpath = np.zeros([(l_p - 1) * (1 + resolution) + 1, dim])
        for step_id in range(l_p - 1):
            for i_res in range(resolution + 1):
                scanpath[step_id * (resolution + 1) + i_res, :] = path[
                    step_id, :
                ] + (path[step_id + 1, :] - path[step_id, :]) * (i_res) / (
                    resolution + 1
                )
        scanpath[-1] = path[-1]
    if mode == "compressed":
        scanpath = np.zeros(shape=path.shape)
        scanpath[0, :] = path[0, :]
        cp = 1
        csp = 1
        acc_path = 0
        while cp < l_p:
            acc_path += np.linalg.norm(path[cp - 1, :] - path[cp, :])
            if acc_path >= stepsize:
                scanpath[csp, :] = path[cp, :]
                cp += 1
                csp += 1
                acc_path = 0
            else:
                cp += 1
        scanpath = scanpath[:csp, :]
    if mode == "segmented":
        total_path = 0
        for step_id in range(l_p - 1):
            total_path += np.linalg.norm(
                path[step_id, :] - path[step_id + 1, :]
            )
        stepsize = total_path / (resolution - 1)

        scanpath = np.zeros(shape=path.shape)
        scanpath[0, :] = path[0, :]
        cp = 1
        csp = 1
        acc_path = 0
        while cp < l_p:
            if (
                acc_path + np.linalg.norm(path[cp - 1, :] - path[cp, :])
                >= stepsize
            ):
                reststep = stepsize - acc_path
                scanpath[csp, :] = path[cp, :]
                cp += 1
                csp += 1
                acc_path = reststep
            else:
                cp += 1
        scanpath = scanpath[:csp, :]

    pass



"""This module collects some examples of loss functions,
as well as a simple gradient descent function."""

from typing import Callable

import numpy as np
from qiskit import (
    ClassicalRegister,
    QuantumCircuit,
    QuantumRegister,
)
from qiskit.providers.basic_provider import BasicProvider

from hyvis.dr_tools import AffineSubspace, numeric_gradient


def gradient_descent_training(
    func: Callable[[np.ndarray], float],
    subspace: AffineSubspace,
    rate: float,
    steps: int,
    epsilon: float,
):
    """This function caclulates a training trajectory using gradient descent.

    Input:
    - func: any function mapping an array of real numbers to a real number
    - subspace: an affine linear subspace of the parameter space,
        its center is the initial state of the trainging and
        its directions are the allowed directions in the training
    - rate: the learning rate, ie the factor put on the gradient in each step
    - steps: the number of steps taken, including the initial step
    - epsilon: the offset used in the approximation, should either be small
        or chosen according to some parameter shift rule

    """

    dim = subspace.center.shape[1]

    traj = np.ndarray(shape=(steps, dim))

    # taking the negative of the loss landscape to get actual descent
    def desfunc(arg):
        return -func(arg)

    for step_id in range(steps):
        traj[step_id, :] = subspace.center
        subspace.center = subspace.center + rate * numeric_gradient(
            func=desfunc, subspace=subspace, epsilon=epsilon, relative=True
        )

    return traj


def relative_entropy_univariate_gaussians(
    par: np.ndarray,
) -> float:
    """Calculates the KL divergence of two univariate Gaussians,
    given their mean and variance."""

    par = par.flatten()

    if not par.shape == (4,) or not par[1] > 0 or not par[3] > 0:
        raise AttributeError(
            "Parameters must be given as np.array([mu1, sigma1^2, mu2, sigma2^2])."
        )

    return (
        (par[1] / par[3])
        + (np.power((par[0] - par[2]), 2) / par[3])
        - np.log((par[1] / par[3]))
        - 1
    ) / 2


def relative_entropy_univariate_gaussians_logscale_variance(
    par: np.ndarray,
) -> float:
    """Calculates the KL divergence of two univariate Gaussians,
    given their mean and variance.
    This version takes the exponential of the parameter corresponding
    to the variance, mostly for ease of plotting."""

    par = par.flatten()

    par[1] = np.exp(par[1])
    par[3] = np.exp(par[3])

    if not par.shape == (4,) or not par[1] > 0 or not par[3] > 0:
        raise AttributeError(
            "Parameters must be given as np.array([mu1, sigma1^2, mu2, sigma2^2])."
        )

    return (
        (par[1] / par[3])
        + (np.power((par[0] - par[2]), 2) / par[3])
        - np.log((par[1] / par[3]))
        - 1
    ) / 2


def relative_entropy_multivariate_gaussians(
    mu1: np.ndarray,
    sigma1: np.ndarray,
    mu2: np.ndarray,
    sigma2: np.ndarray,
):
    """Calculates the KL divergence of two multivariate Gaussians
    given theirs means and covariance matrices. The latter have to be positive definite.
    """
    k = mu1.shape[0]

    if (
        (not mu1.shape == (k, 1))
        or (not mu2.shape == (k, 1))
        or (not sigma1.shape == (k, k))
        or (not sigma2.shape == (k, k))
    ):
        raise AttributeError("Inconsistent dimensions.")

    if np.linalg.det(sigma1) > 0 and np.linalg.det(sigma2) > 0:
        return (
            np.trace(np.dot(np.linalg.inv(sigma2), sigma1))
            + np.dot(
                np.transpose(mu2 - mu1),
                np.dot(np.linalg.inv(sigma2), (mu2 - mu1)),
            )
            + np.log(np.linalg.det(sigma2) / np.linalg.det(sigma1))
            - k
        ).item() / 2
    else:
        return np.nan


def relative_entropy_multivariate_gaussians_unified(
    par: np.ndarray,
):
    """Calculates the KL divergence of two multivariate Gaussians.

    This version takes a flat array and then constructs
    the means and covariance matrices from it, the latter should be positive definite.
    """
    par = par.flatten()

    l_par = len(par)

    found_k = False
    for k in range(int(np.ceil(np.sqrt(l_par / 2)))):
        if l_par / 2 == k * (1 + k):
            found_k = True
            break

    if found_k is False:
        raise AttributeError(
            "Invalid number of entries for two multivariate Gaussians."
        )

    mu1 = par[0:k].reshape((k, 1))
    sigma1 = par[k : (k + k * k)].reshape((k, k))
    mu2 = par[(k + k * k) : (2 * k + k * k)].reshape((k, 1))
    sigma2 = par[(2 * k + k * k) : (2 * k + 2 * k * k)].reshape((k, k))

    return relative_entropy_multivariate_gaussians(mu1, sigma1, mu2, sigma2)


# the following are building blocks for a landscape corresponding to the maxcut problem
# of graphs using a quantum circuit


def qaoa_circuit(
    n_vertices: int,
) -> tuple[QuantumCircuit, QuantumRegister, ClassicalRegister]:
    vertices_r = QuantumRegister(n_vertices, name="v")
    classic_r = ClassicalRegister(n_vertices, name="output")
    qc = QuantumCircuit(vertices_r, classic_r)
    return qc, vertices_r, classic_r


def initialize_circuit(
    qc: QuantumCircuit, vertices_r: QuantumRegister
) -> QuantumCircuit:
    for i in range(len(vertices_r)):
        qc.h(vertices_r[i])
    return qc


def u_c(gamma, qc, vertices_r, w):
    for i in range(len(w)):
        for j in range(i + 1, len(w)):
            if w[i, j] != 0:
                qc.cx(vertices_r[i], vertices_r[j])
                qc.rz(gamma * w[i, j], vertices_r[j])
                qc.cx(vertices_r[i], vertices_r[j])
    return qc


def u_m(beta, qc, vertices_r):
    for i in range(len(vertices_r)):
        qc.rx(2 * beta, vertices_r[i])
    return qc


def qaoa_circuit_with_layers(w, gamma_list, beta_list):
    n_vertices = len(w)
    qc, vertices_r, classic_r = qaoa_circuit(n_vertices)
    qc = initialize_circuit(qc, vertices_r)
    for gamma, beta in zip(gamma_list, beta_list):
        qc = u_c(gamma, qc, vertices_r, w)
        qc.barrier()
        qc = u_m(beta, qc, vertices_r)
    qc.measure(vertices_r, classic_r)
    return qc


def maxcut_landscape_qaoa(parameters: list, w: np.ndarray) -> float:
    """Evaluate the MaxCut objective function for given parameters
    using a qoao circuit.

    Input:

        parameters: list of [gamma_1, ... , gamma_p, beta_q, ... ,beta_p]
            that defines the angles used in the trotterized adiabatic process
            of qaoa

        w: weight matrix of a graph

    Output:

        negative of the average cut value obtained by the given qaoa parameters

    """

    nodes = w.shape[0]
    p = len(parameters) // 2
    gamma_list = parameters[:p]
    beta_list = parameters[p:]
    qc = qaoa_circuit_with_layers(w, gamma_list, beta_list)

    backend = BasicProvider().get_backend("basic_simulator")
    job = backend.run(qc, shots=1024)
    counts = job.result().get_counts()

    obj = 0
    for bitstring, count in counts.items():
        z = [1 if bit == "1" else -1 for bit in bitstring]
        obj += (
            sum(
                w[i, j] * (1 - z[i] * z[j]) / 2
                for i in range(nodes)
                for j in range(i + 1, nodes)
            )
            * count
        )

    return -obj / 1024



"""This is a module for creating videos of scans of a function landscape."""

from typing import Callable, List, Optional, Union
from warnings import warn

import numpy as np
from matplotlib import animation
from matplotlib import colors as mcolors
from matplotlib import pyplot as plt
from matplotlib.collections import LineCollection
from sklearn.decomposition import PCA

from .basic_scans import (
    ScanCollection,
    collective_scan_linear,
    hessian_scan,
    landscape_scan_linear,
)
from .dr_tools import AffineSubspace, numeric_hessian, subspace_projection


class VideoScan:
    """This object contains data corresponding to linear scans
    that form the frames of a video.

        Attributes:

            result: a grid of values sampled from a real valued function.
                the last dimension enumerates the frames

            subspaces: A list of affine linear subspaces
                along which the samples were taken.

            scope: tells you how far the subspace was scanned in each direction.
                shape is (number of directions, 2)

            func: optional attribute to record the function that was sampled.


        Methods:

            animate: creates and plays the actual video based on the results.
                this can take quite long
    """

    def __init__(
        self,
        result: np.ndarray,
        subspaces: List[AffineSubspace],
        scope: np.ndarray,  # shape (number of directions, 2). records how far
        # the landscape was scanned in each direction
        func: Optional[Callable[[np.ndarray], float]] = None,
        trajectory: Optional[np.ndarray] = None,
    ):
        self.result = result
        self.subspaces = subspaces
        self.scope = scope
        self.func = func
        self.trajectory = trajectory

    def animate(
        self,
        xlabel: Optional[str] = None,
        ylabel: Optional[str] = None,
        title: Optional[str] = None,
        show_trajectory: Optional[bool] = True,
        trajectory_fade: Optional[bool] = True,
        trajectory_color: Optional[str] = "red",
        **plot_kwargs,
    ):
        """This method creates an animation of the video.
        By default this includes projections of the remaining trajectory onto each
        frame. Note that if the range for the colors is not fixed (e.b. by setting
        vmin and vmax in plot_kwargs) then each frame will be colored separately.

        Input:
            - show_trajectory: whether to include the lineplot
                that is the remaining trajectory
            - trajectory_fade: wheter to have the line fade
                towards the end of the trajectory
            - trajectory_color: the color of the lineplot

        Output:
            None, it immediately opens a video player when used in a notebook

        """

        if "vmin" not in plot_kwargs or "vmax" not in plot_kwargs:
            warn(
                """The range of the colormap has not been specified, each frame will be colored individually. To get consistent coloring specify vmin and vmax as arguments for this method."""
            )

        plt.rcParams["animation.html"] = "jshtml"
        plt.ioff()

        fig, ax = plt.subplots()

        def create_animation(t):
            plt.cla()
            step_num = len(self.subspaces)

            # showing the scan
            plt.imshow(
                np.transpose(self.result[:, :, t]),
                extent=[
                    self.scope[0, 0],
                    self.scope[0, 1],
                    self.scope[1, 0],
                    self.scope[1, 1],
                ],
                **plot_kwargs,
            )
            plt.xlabel(xlabel)
            plt.ylabel(ylabel)
            plt.title(title)

            if self.trajectory is not None:
                if show_trajectory is True:
                    # getting the projected trajectory
                    coeff = np.zeros([step_num - t, 2])
                    for step_id in range(t, step_num):
                        coeff[step_id - t, :] = np.linalg.lstsq(
                            self.subspaces[t].directions.transpose(),
                            (
                                self.trajectory[step_id, :]
                                - self.subspaces[t].center
                            ).transpose(),
                            rcond=None,
                        )[0].flatten()

                    x = coeff[:, 0]
                    y = coeff[:, 1]
                    reststeps = x.shape[0]

                    colors = [mcolors.to_rgba(trajectory_color)] * reststeps
                    if trajectory_fade is True:
                        fade = (
                            1
                            + np.cos(
                                np.pi
                                * np.linspace(0, reststeps - 1, reststeps)
                                / (reststeps)
                            )
                        ) / 2
                        for step_id in range(reststeps):
                            step_color = list(colors[step_id])
                            step_color[3] = fade[step_id]
                            colors[step_id] = tuple(step_color)

                    points = np.array([x, y]).T.reshape(-1, 1, 2)
                    segments = np.concatenate(
                        [points[:-1], points[1:]], axis=1
                    )
                    lc = LineCollection(segments, colors=colors)
                    ax.add_collection(lc, autolim=False)

        frames = len(self.subspaces)
        ani = animation.FuncAnimation(fig, create_animation, frames=frames)

        return ani


class VideoCollectiveScan:
    """This object contains data corresponding to collective scans
    that form the frames of a video.

        Attributes:

            result: a grid of values sampled from a real valued function.
                the last dimension enumerates the frames

            subspaces: A list of affine linear subspaces
                along which the samples were taken.

            scope: tells you how far the subspace was scanned in each direction.
                shape is (number of directions, 2)

            func: optional attribute to record the function that was sampled.


        Methods:

            animate: creates and plays the actual video based on the results.
                this can take quite long
    """

    def __init__(
        self,
        scans: List[ScanCollection],
        scope: np.ndarray,  # shape (number of directions, 2). records how far
        # the landscape was scanned in each direction
        func: Optional[Callable[[np.ndarray], float]] = None,
        trajectory: Optional[np.ndarray] = None,
    ):
        self.scans = scans
        self.scope = scope
        self.func = func
        self.trajectory = trajectory

    def animate(self, **plot_kwargs):
        """This method creates an animation of the video.
        By default this includes projections of the remaining trajectory onto each
        frame.

        Input:
            - show_trajectory: whether to include the lineplot
                that is the remaining trajectory
            - trajectory_fade: wheter to have the line fade
                towards the end of the trajectory
            - trajectory_color: the color of the lineplot

        Output:
            None, it immediately opens a video player when used in a notebook

        """

        plt.rcParams["animation.html"] = "jshtml"
        plt.ioff()

        fig, ax = plt.subplots()

        def create_animation(t):
            plt.cla()
            # step_num = len(self.scans)

            # showing the scan
            self.scans[t].show(**plot_kwargs)

            ax = plt.gca()
            yabs_max = abs(max(ax.get_ylim(), key=abs))
            ax.set_ylim(ymin=-yabs_max, ymax=yabs_max)

        frames = len(self.scans)
        ani = animation.FuncAnimation(fig, create_animation, frames=frames)

        return ani


def volume_scan(
    func: Callable[[np.ndarray], float],
    subspace: AffineSubspace,
    scope: Optional[Union[np.ndarray, float]] = 5,
    resolution: Optional[Union[int, np.ndarray]] = 10,
    pools: Optional[int] = 1,
) -> VideoScan:
    """This function creates a videoscan where each frame scans the subspace except
    in the last direction, which is instead used to move the space. For example,
    if the subspace is 3D this will create a video of 2D scans.

    The subspace should have at least 3 dimensions. It can have more, but then
    it can not be animated.
    """

    d_num = subspace.directions.shape[0]

    if not isinstance(scope, np.ndarray):
        scope = scope * np.append(-np.ones([d_num, 1]), np.ones([d_num, 1]), 1)

    if not (scope[:, 0] < scope[:, 1]).all():
        raise ValueError(
            """scope[id_d,0] must be strictly smaller than scope[id_d,1] for each
            direction."""
        )

    if np.isscalar(resolution):
        resolution = resolution * np.ones(d_num, dtype=int)

    scanres = resolution[-1]
    subspaces = []
    for i_scan in range(scanres):
        center = np.array(
            [
                (
                    subspace.center.flatten()
                    + scope[-1, 0] * subspace.directions[-1, :]
                    + (-scope[-1, 0] + scope[-1, 1])
                    * i_scan
                    * ((-scope[-1, 0] + scope[-1, 1]) / (scanres - 1))
                    * subspace.directions[-1, :]
                )
            ]
        )
        subspaces.append(
            AffineSubspace(subspace.directions[:-1, :], center=center)
        )

    scan = landscape_scan_linear(
        func=func,
        subspace=subspace,
        scope=scope,
        resolution=resolution,
        pools=pools,
    )

    return VideoScan(
        result=scan.result,
        subspaces=subspaces,
        scope=scan.scope,
        func=func,
    )


def operator_scan(
    func: Callable[[np.ndarray], float],
    subspace: AffineSubspace,
    operator: Optional[np.ndarray] = None,
    shift: Optional[np.ndarray] = None,
    steps: Optional[int] = 10,
    scope: Optional[Union[np.ndarray, float]] = 5,
    resolution: Optional[Union[int, np.ndarray]] = 10,
    pools: Optional[int] = 1,
) -> VideoScan:
    """performs some number of scans, where in each step the
    subspace is transformed by an affine linear operator"""

    d_dim = subspace.directions.shape[1]
    d_num = subspace.directions.shape[0]

    if not isinstance(scope, np.ndarray):
        scope = scope * np.append(-np.ones([d_num, 1]), np.ones([d_num, 1]), 1)

    if not (scope[:, 0] < scope[:, 1]).all():
        raise ValueError(
            """scope[id_d,0] must be strictly smaller than scope[id_d,1] for each
            direction."""
        )

    if np.isscalar(resolution):
        resolution = resolution * np.ones(d_num, dtype=int)

    if operator is None:
        operator = np.eye(d_dim)
    if shift is None:
        shift = np.zeros(shape=subspace.center.shape)

    result = np.zeros(shape=np.append(resolution, steps))

    subspaces = []
    for i_step in range(steps):
        scan = landscape_scan_linear(
            func, subspace, scope, resolution, pools=pools
        )
        subspaces.append(subspace)
        result[..., i_step] = scan.result

        directions = (operator @ subspace.directions.T).T
        center = subspace.center + shift
        subspace = AffineSubspace(directions=directions, center=center)

    return VideoScan(
        result=result,
        subspaces=subspaces,
        scope=scope,
        func=func,
    )


def trajectory_scan_stepwise_pca(
    func: Callable[[np.ndarray], float],
    trajectory: np.ndarray,
    scope: Union[np.ndarray, float] = 5,
    resolution: Optional[Union[int, np.ndarray]] = 10,
    pools: Optional[int] = 1,
) -> VideoScan:
    """The purpose of this function is to scan a landscape defined by func
    along the path defined by trajectory. For each step in the trajectory it will
    project the remaining trajectory onto the orthogonal complement of the gradient
    of the previous step and then do pca on that, to get the scan directions.

    The first and the last frame show the pca of the entire trajectory without any
    projections because above described method cannot be applied there.

    Input:

        func: The function that defines the landscape.

        trajectory: an array of points in the lanscape
            must be of shape (number of steps, dimension of superspace)

        scope: How far to scan in each direction of subspace.
        The size has to be (d_num,2), where for each direction the first entry is the
        beginning of the scope and the second is the end.
        If provided as float, the beginning will be -scope and the end +scope.

        resolution: How many samples to take in each direction of subspace.
        If provided as int, the resolution will be the same for each direction.

    Output:

        A trajectory scan object.

    """

    d_num = 2
    d_dim = trajectory.shape[1]

    if not isinstance(scope, np.ndarray):
        scope = scope * np.append(-np.ones([d_num, 1]), np.ones([d_num, 1]), 1)

    if isinstance(resolution, int):
        resolution = resolution * np.ones(d_num, dtype=int)

    traj_length = trajectory.shape[0]
    result = np.zeros(shape=np.append(resolution, traj_length))
    subspaces = [None] * traj_length
    directions = np.zeros([d_num, d_dim])

    # doing the first scan separately because there is no preceding step to define the
    # direction. instead it will just show the pca of the whole trajectory
    step_id = 0

    directions = PCA(n_components=2).fit(trajectory).components_
    directions = np.dot(np.dot(directions, directions.transpose()), directions)

    center = PCA(n_components=2).fit(trajectory).mean_.reshape([1, d_dim])

    subspace = AffineSubspace(directions=directions, center=center)

    scan = landscape_scan_linear(
        func, subspace, scope, resolution, pools=pools
    )
    result[:, :, step_id] = scan.result
    subspaces[step_id] = subspace

    # doing the stepwise pca scan for the remaining steps
    for step_id in range(1, traj_length - 1):
        stepspace = AffineSubspace(
            directions=np.array(
                [
                    (trajectory[step_id, :] - trajectory[step_id - 1, :])
                    / np.linalg.norm(
                        trajectory[step_id, :] - trajectory[step_id - 1, :]
                    )
                ]
            ),
            center=np.array([trajectory[step_id, :]]),
        )
        cloud = subspace_projection(
            trajectory[step_id:, :], stepspace, orthogonal=True, relative=True
        )
        directions_old = directions
        directions = PCA(n_components=2).fit(cloud).components_
        directions = np.dot(
            np.dot(directions, directions.transpose()), directions
        )
        # now making sure that directions dont 'flip sign' by matching them with
        for d_id in range(d_num):
            if np.linalg.norm(
                directions[d_id, :] - directions_old[d_id, :]
            ) > np.linalg.norm(-directions[d_id, :] - directions_old[d_id, :]):
                directions[d_id, :] = -directions[d_id, :]

        subspace = AffineSubspace(
            directions=directions, center=np.array([trajectory[step_id, :]])
        )
        scan = landscape_scan_linear(
            func, subspace, scope, resolution, pools=pools
        )

        result[:, :, step_id] = scan.result
        subspaces[step_id] = subspace

    # doing the last scan separately because there are not enough following points
    # to do pca on, instead it will just show the pca of the whole trajectory
    step_id = traj_length - 1

    directions = PCA(n_components=2).fit(trajectory).components_
    directions = np.dot(np.dot(directions, directions.transpose()), directions)

    center = PCA(n_components=2).fit(trajectory).mean_.reshape([1, d_dim])

    subspace = AffineSubspace(directions=directions, center=center)

    scan = landscape_scan_linear(
        func, subspace, scope, resolution, pools=pools
    )
    result[:, :, step_id] = scan.result
    subspaces[step_id] = subspace

    return VideoScan(
        result=result,
        subspaces=subspaces,
        scope=scope,
        func=func,
        trajectory=trajectory,
    )


def trajectory_scan_stepwise_hessian(
    func: Callable[[np.ndarray], float],
    trajectory: np.ndarray,
    scope: Union[np.ndarray, float] = 5,
    resolution: Optional[Union[int, np.ndarray]] = 10,
    epsilon: Optional[float] = 0.01,
    sharpness: Optional[int] = 2,
    pools: Optional[int] = 1,
) -> VideoCollectiveScan:
    """This function creates one frame per point in the trajectory by performing
    a hessian_scan on the orthogonal complement of the
    direction of the next step.
    If the latter is the same a the gradient, then this is related to the
    second fundamental form.

    Note: since this cannot be done with the last frame,
    it shows the hessian scan on the full space.

    Input:

        func: The function that defines the landscape.

        trajectory: an array of points in the landscape
            must be of shape (number of steps, dimension of superspace)

        scope: How far to scan in each direction of subspace.
        The size has to be (d_num,2), where for each direction the first entry is the
        beginning of the scope and the second is the end.
        If provided as float, the beginning will be -scope and the end +scope.

        resolution: How many samples to take in each direction of subspace.
        If provided as int, the resolution will be the same for each direction.

    Output:

        A trajectory scan object.

    """

    d_dim = trajectory.shape[1]
    d_num = d_dim - 1

    if not isinstance(scope, np.ndarray):
        scope = scope * np.append(-np.ones([d_num, 1]), np.ones([d_num, 1]), 1)

    if isinstance(resolution, int):
        resolution = resolution * np.ones(d_num, dtype=int)

    traj_length = trajectory.shape[0]

    # result = np.zeros(shape=np.append(resolution, traj_length))
    scans = [None] * traj_length
    directions = np.zeros([d_num, d_dim])

    # doing the stepwise pca scan for the remaining steps
    for step_id in range(0, traj_length - 1):
        # getting direction of next step
        stepspace = AffineSubspace(
            directions=np.array(
                [
                    (trajectory[step_id + 1, :] - trajectory[step_id, :])
                    / np.linalg.norm(
                        trajectory[step_id + 1, :] - trajectory[step_id, :]
                    )
                ]
            ),
            center=np.array([trajectory[step_id, :]]),
            sharpness=sharpness,
        )

        stepspace_oc = stepspace.orth()
        H = numeric_hessian(func=func, subspace=stepspace_oc, epsilon=epsilon)
        H.calc_evs()

        directions_old = directions
        directions = np.dot(
            H.eigenvectors.transpose(), stepspace_oc.directions
        )
        # directions = stepspace_oc.directions
        # now making sure that directions dont 'flip sign' by matching them with
        # the previous step
        for d_id in range(d_num):
            if np.linalg.norm(
                directions[d_id, :] - directions_old[d_id, :]
            ) > np.linalg.norm(-directions[d_id, :] - directions_old[d_id, :]):
                directions[d_id, :] = -directions[d_id, :]

        stepspace_h = AffineSubspace(
            directions=directions,
            center=stepspace_oc.center,
            orthonormalize=False,
        )

        scans[step_id] = collective_scan_linear(
            func=func,
            subspace=stepspace_h,
            scope=scope,
            resolution=resolution,
            pools=pools,
        )

    # doing the last step separately
    step_id = traj_length - 1
    finalspace = AffineSubspace(
        directions=np.eye(d_dim),
        center=trajectory[[step_id], :],
        sharpness=stepspace.sharpness,
    )
    scans[step_id] = hessian_scan(
        func=func,
        subspace=finalspace,
        scope=np.append(scope, scope[[d_dim - 2], :], axis=0),
        resolution=np.append(resolution, resolution[d_dim - 2]),
        epsilon=epsilon,
        pools=pools,
    )[0]

    return VideoCollectiveScan(
        scans=scans, scope=scope, func=func, trajectory=trajectory
    )
